"""Implements an interface for backtesting."""
import datetime
import json
import os

import bson
from IPython.display import display, clear_output
from ipywidgets import FloatProgress
from pymongo import MongoClient

from ia.gaius.data_ops import Data
from ia.gaius.tests import classification, utility

import plotly.express as px
from plotly.subplots import make_subplots
import plotly.graph_objects as go
import json 
from math import pi
import pandas as pd 
from sklearn.metrics import accuracy_score, precision_score

class BackTest:
    """Provides an interface for backtesting."""
    def __init__(self, **kwargs):
        """
        Pass this a configuration dictionary as the argument such as:

        test_results_database = 'mongodb://mongo-kb:27017'

        test_config = {'name':'backtest-classification',
                        'test_type': 'classification', ## 'classification' or 'utility'
                        'utype': None, ## If 'test_type' = 'utility', then set to either 'polarity' or 'value'
                        'shuffle_data': True,
                        'fresh_start_memory': True, ## Clear all memory when starting and between runs
                        'mongo_location': test_results_database, ## location of mongo db where test results will be stored
                        'learning_strategy': 'continuous', # None, 'continuous' or 'on_error'
                        'bottle': bottle,
                        'data_source': dataset, ## or provide 'data_directories' as iterable of data.
                        'percent_reserved_for_training': 20,
                        'percent_of_dataset_chosen': 100,
                        'total_test_counts': 1 }

        test = BackTest(**test_config)

        mongo_location provides the location of a mongo database where the test results will be stored.

        Option of either data_source or data_directories can be provided:

            data_source provided is a sequence of sequences of GDF objects.
            data_directories provided should be a list of directories containing files of GDF as json dumps.

        """
        self.configuration = kwargs
        self.errors = []
        self.name = str(kwargs['name'])
        self.test_type = kwargs['test_type']
        self.utype = kwargs['utype']
        self.shuffle_data = kwargs['shuffle_data']
        self.learning_strategy = kwargs['learning_strategy']
        self.fresh_start_memory = kwargs['fresh_start_memory']
        self.mongo_location = kwargs['mongo_location']
        self.bottle = kwargs['bottle']
        self.bottle.summarize_for_single_node = False
        if 'data_directories' in self.configuration:
            self.data_directories = kwargs['data_directories']
            self.data_source = None
        elif 'data_source' in self.configuration:
            self.data_source = kwargs['data_source']
            self.configuration['data_source'] = 'from-source'
            self.data_directories = None
        self.percent_reserved_for_training = int(kwargs['percent_reserved_for_training'])
        self.percent_of_dataset_chosen = int(kwargs['percent_of_dataset_chosen'])
        self.total_test_counts = int(kwargs['total_test_counts'])
        self.current_test_count = 0

        self.mongo_client = MongoClient(self.mongo_location)  # , document_class=OrderedDict)
        # Collection storing the backtesting results is the bottle's name.
        self.mongo_client.backtesting = self.mongo_client['{}-{}-{}'.format(
            self.name, self.bottle.genome.agent, self.bottle.name)]
        self.test_configuration = self.mongo_client.backtesting.test_configuration
        self.test_status = self.mongo_client.backtesting.test_status
        self.test_errors = self.mongo_client.backtesting.test_errors
        self.backtesting_log = self.mongo_client.backtesting.backtesting_log
        self.interrupt_status = self.mongo_client.backtesting.interrupt_status

        if self.test_type == "utility":
            self._tester = utility.Tester(**kwargs)
        elif self.test_type == 'classification':
            self._tester = classification.Tester(**kwargs)

        if self.data_directories:
            self.data = Data(data_directories=self.data_directories)
        elif self.data_source:
            self.data = Data(dataset=self.data_source)
        self.data.prep(self.percent_of_dataset_chosen, self.percent_reserved_for_training, shuffle=self.shuffle_data)
        sequence_count = len(self.data.train_sequences) + len(self.data.test_sequences)

        self.number_of_things_to_do = self.total_test_counts * sequence_count
        self.number_of_things_done = 0
        self.status = "not started"

        self.interrupt_status.replace_one({}, {'interrupt': False}, upsert=True)
        self.test_status.replace_one({}, {'status': 'not started',
                                          'number_of_things_to_do': self.total_test_counts * sequence_count,
                                          'number_of_things_done': 0,
                                          'current_test_count': self.current_test_count
                                          }, upsert=True)

        self.test_errors.drop()
        self.backtesting_log.drop()
        self.backtesting_log.insert_one({"timestamp_utc": datetime.datetime.utcnow(), "status": "test-started"})

        self.test_configuration.drop()
        self.test_configuration.insert_one({
            'name': self.name,
            'test_type': self.test_type,
            'utype': self.utype,
            'shuffle_data': self.shuffle_data,
            'learning_strategy': self.learning_strategy,
            'fresh_start_memory': self.fresh_start_memory,
            "bottle_name": self.bottle.name,
            "agent": self.bottle.genome.agent,
            "ingress_nodes": self.bottle.ingress_nodes,
            "query_nodes": self.bottle.query_nodes
        })

        headers = ["Test Run", "Trial", "Phase", "Filename", "Historical"] + [node["name"] for node in
                                                                              self.bottle.query_nodes] + ["hive"]
        self.backtesting_log.insert_one({"headers": headers})
        self.progress = FloatProgress(min=0, max=self.number_of_things_to_do, description="Starting...", bar_style="info")

        print("Recording results at '%s-%s-%s'" % (self.name, self.bottle.genome.agent, self.bottle.name))

    def _reset_test(self):
        """Reset the instance to the state it was in when created."""
        self.backtesting_log.insert_one({"timestamp_utc": datetime.datetime.utcnow(), "status": "resetTest"})
        if self.data_directories:
            self.data = Data(data_directories=self.data_directories)
        elif self.data_source:
            self.data = Data(dataset=self.data_source)
        self.data.prep(self.percent_of_dataset_chosen, self.percent_reserved_for_training, shuffle=self.shuffle_data)
        self._tester.next_test_prep()

    def _end_test(self):
        """Called when the test ends."""
        self.status = "finished"
        self.backtesting_log.insert_one({"timestamp_utc": datetime.datetime.utcnow(), "status": "test-ended"})
        nodes_status = self.bottle.show_status()
        self.test_status.replace_one({}, {'status': 'finished',
                                          'nodes_status': nodes_status,
                                          'number_of_things_to_do': self.number_of_things_to_do,
                                          'number_of_things_done': self.number_of_things_to_do,
                                          'current_test_count': self.current_test_count}, upsert=True)

    def run(self):
        display(self.progress)
        self.backtesting_log.insert_one({"timestamp_utc": datetime.datetime.utcnow(), "status": "run"})
        while self.current_test_count < self.total_test_counts:
            self.current_test_count += 1
            self._setup_training()
            for sequence in self.data.train_sequences:
                self._train(sequence)
                self.number_of_things_done += 1
                self.progress.value = self.number_of_things_done
                self.progress.description = '%0.2f%%' % (100 * self.number_of_things_done / self.number_of_things_to_do)

            self._setup_testing()
            for sequence in self.data.test_sequences:
                self._test(sequence)
                self.number_of_things_done += 1
                self.progress.value = self.number_of_things_done
                self.progress.description = '%0.2f%%' % (100 * self.number_of_things_done / self.number_of_things_to_do)

            if self.current_test_count < self.total_test_counts:
                self._reset_test()
            else:
                self._end_test()
                clear_output()

    def _setup_training(self):
        """Setup instance for training."""
        self.backtesting_log.insert_one({"timestamp_utc": datetime.datetime.utcnow(), "status": "setupTraining"})
        self.status = "training"
        self.test_status.replace_one({}, {'status': 'training',
                                          'number_of_things_to_do': self.number_of_things_to_do,
                                          'number_of_things_done': self.number_of_things_done,
                                          'current_test_count': self.current_test_count}, upsert=True)
        if self.fresh_start_memory:
            self.bottle.clear_all_memory()
        return 'ready'

    def _train(self, sequence):
        """Train with the sequence in *sequence*."""
        # get a sequence either from a file, or directly as a list:
        if self.data_directories:
            self.backtesting_log.insert_one(
                {"timestamp_utc": datetime.datetime.utcnow(), "status": "training", "file": os.path.basename(sequence)})
            with open(sequence) as f:
                sequence = [json.loads(data.strip()) for data in f if data]
        elif self.data_source:
            sequence = sequence
            self.backtesting_log.insert_one({"timestamp_utc": datetime.datetime.utcnow(), "status": "training"})

        ## Train the sequence:
        result_log_record = self._tester.train(sequence)
        result_log_record['trial'] = self.number_of_things_done
        result_log_record['run'] = self.current_test_count
        self.backtesting_log.insert_one(result_log_record)
        return 'ready'

    def _setup_testing(self):
        """Set up the instance to begin backtesting."""
        self.backtesting_log.insert_one({"timestamp_utc": datetime.datetime.utcnow(), "status": "setupTesting"})
        self.status = "testing"
        self.test_status.replace_one({}, {'status': 'testing',
                                          'number_of_things_to_do': self.number_of_things_to_do,
                                          'number_of_things_done': self.number_of_things_done,
                                          'current_test_count': self.current_test_count}, upsert=True)
        return 'ready'

    def _test(self, sequence):
        """Run the backtest on *sequence*."""
        ## get a sequence either from a file, or directly as a list:
        if self.data_directories:
            self.backtesting_log.insert_one(
                {"timestamp_utc": datetime.datetime.utcnow(), "status": "testing", "file": os.path.basename(sequence)})
            with open(sequence) as f:
                sequence = [json.loads(data.strip()) for data in f if data]
        elif self.data_source:
            sequence = sequence
            self.backtesting_log.insert_one({"timestamp_utc": datetime.datetime.utcnow(), "status": "testing"})

        ## Test the sequence and record the results.
        result_log_record = self._tester.test(sequence)
        result_log_record['trial'] = self.number_of_things_done
        result_log_record['run'] = self.current_test_count
        self.backtesting_log.insert_one(bson.son.SON(result_log_record))
        return 'ready'


def classification_report(back_tester : BackTest): 
    class_values = get_class_distribution(back_tester)

    plot_class_distributions(back_tester, class_values)

    predictor_operator_characteristic(back_tester, class_values)

    fidelity_charts(back_tester)

def get_class_distribution(back_tester : BackTest) -> dict: 
    return_dict = {"training" : {}, "testing" : {}} 
    training_class_labels = [] 
    testing_class_labels = [] 
    if back_tester.data_directories != None: 
        for file in back_tester.data.train_sequences: 
            with open(file) as f: 
                sequence = [json.loads(data.strip()) for data in f if data] 
                if(sequence[-1]["strings"][-1] != ""): 
                    training_class_labels.append(sequence[-1]["strings"][-1])

        for key in set(training_class_labels):
            return_dict["training"][key] = training_class_labels.count(key)


        for file in back_tester.data.test_sequences:
            with open(file) as f:
                sequence = [json.loads(data.strip()) for data in f if data]
                if(sequence[-1]["strings"][-1] != ""):
                    testing_class_labels.append(sequence[-1]["strings"][-1])

        for key in set(testing_class_labels):
            return_dict["testing"][key] = testing_class_labels.count(key)

    else:
        for seq in back_tester.train_sequences:
            if(sequence[-1]["strings"][-1] != ""):
                training_class_labels.append(sequence[-1]["strings"][-1])

        for key in set(training_class_labels):
            return_dict["training"][key] = training_class_labels.count(key)

        for seq in back_tester.test_sequences:
            if(sequence[-1]["strings"][-1] != ""):
                testing_class_labels.append(sequence[-1]["strings"][-1])

        for key in set(testing_class_labels):
            return_dict["testing"][key] = testing_class_labels.count(key)

    return return_dict

def plot_class_distributions(back_tester : BackTest, class_values): 
    # make subplot figure 
    fig = make_subplots(rows=1, cols=2, specs=[[{'type':'domain'}, {'type':'domain'}]])

    # get training data distribution
    x = class_values["training"]
    data = pd.Series(x).reset_index(name='value').rename(columns={'index': 'class'})

    # add first pie subplot
    fig.add_trace(
        go.Pie(labels=data["class"].to_list(), values=data["value"].to_list(), name="Training", 
               title='Training', texttemplate = "%{label}<br>(%{percent:.2f})"),
        1, 1)

    fig.update_traces(hole=.4, hoverinfo="label+percent+name")

    # get testing data distribution
    x = class_values["testing"]
    data = pd.Series(x).reset_index(name='value').rename(columns={'index': 'class'})

    # add other pie subplot
    fig.add_trace(
        go.Pie(labels=data["class"].to_list(), values=data["value"].to_list(), 
               name="Testing", title='Testing', texttemplate = "%{label}<br>(%{percent:.2f})"),
        1, 2,)

    fig.update_traces(hole=.4, hoverinfo="label+percent+name")

    fig.update_layout(height=500, width=800, title_text="Historical Class Distributions")
    fig.show()

def predictor_operator_characteristic(back_tester : BackTest, class_values): 
    # create some variables to hold data 
    predicted_labels = [] 
    true_labels = [] 
    num_correct = 0 
    num_wrong = 0

    # go through results and record number of correct and incorrect
    for doc in back_tester.backtesting_log.find({"phase" : "testing"}):
        predicted_labels.append(doc['node_predictions']["hive"])
        true_labels.append(doc['historical_expecting'])
        if(doc['historical_expecting'] == doc['node_predictions']["hive"]):
            num_correct += 1
        else:
            num_wrong += 1

    # calculate scores
    accuracy_of_positives = accuracy_score(predicted_labels, true_labels)

    precision_of_positives = num_correct / (num_wrong + num_correct)


    line_color=dict(color="green")

    layout1= go.Layout(title=go.layout.Title(text="Predictor Operator Characteristic",x=0.5),
        xaxis={'title':'Precision of Positives (Positive Predictive Values)','range':[0,1.0]},
        yaxis={'title':'Accuracy of Positives','range':[0,1.0]})

    point_plot=[
      go.Scatter(x=[(1 / len(class_values["testing"].keys())) for i in range(2)],
             y=[i for i in range(2)],
             name="Random Chance",
             legendgroup="Random Chance",
             line=line_color),
        go.Scatter(x = [accuracy_of_positives], y = [precision_of_positives], 
                   name = "GAIUS Performance", legendgroup="GAIUS Performance",
                  marker={"color":"green", "size":25}, mode="markers")
    ]

    go.Figure(data=point_plot, layout=layout1).show()

def fidelity_charts(back_tester : BackTest):
    import collections
    calculated_things = {}
    # get the names of all the nodes, and add hive
    # will probably need to filter some out or not
    nodes = [name["name"] for name in back_tester.bottle.all_nodes]
    nodes.insert(0, "hive")
#     nodes.append("hive")
    
    # set up dictionary, before reading info from backtesting_log
    for node in nodes:
        calculated_things[node] = {}
        calculated_things[node]["node_predictions"] = []
        calculated_things[node]["num_correct"] = 0
        calculated_things[node]["num_wrong"] = 0
        for actual in back_tester.backtesting_log.distinct('historical_expecting'):
            calculated_things[node][actual] = {}
            for predicted in back_tester.backtesting_log.distinct('historical_expecting'):
                calculated_things[node][actual][predicted] = 0
    
    # get information from backesting_log
    for doc in back_tester.backtesting_log.find({"phase" : "testing"}):
        for node in nodes:
            calculated_things[node][doc['historical_expecting']][doc['node_predictions'][node]] += 1
            calculated_things[node]["node_predictions"].append(doc['node_predictions'][node])
            if(doc['historical_expecting'] == doc['node_predictions'][node]):
                calculated_things[node]["num_correct"] += 1
            else:
                calculated_things[node]["num_wrong"] += 1
    
    # get status of bottles
    bottle_statuses = back_tester.bottle.show_status()
    
    # set up suplots
    subplot_spec = []
    subplot_titles = []
    width = 1000
    height = 400 * (len(nodes)+1)
    
    for node in nodes:
        subplot_spec.append([{"type": "table"},
                   {"type": "domain"},
                   {"type": "domain"}])
        
        subplot_titles.append(node.capitalize())
        subplot_titles.append('Fidelity')
        subplot_titles.append("Predicted Class Distributions")
        
    
    subplot_spec.append([{"type": "table", "colspan": 3}, None, None])
    subplot_titles.append("Confusion Matrix")
    
    fig = make_subplots(rows=len(nodes)+1, cols=3,
            specs=subplot_spec,
            subplot_titles=subplot_titles,
            column_widths=[width / 3, width /3, width / 3])
    
    # go through nodes and plot their data
    row_index = 1
    for i, node in enumerate(nodes):
        # make small table with node status
        if(node == "hive"):
            fig.add_trace(go.Table(
                header=dict(values=[node.capitalize() + " Status"],
                    align='left'),
                cells=dict(values=[[""]],
                   align='left')), row_index, 1)
        else:
            fig.add_trace(go.Table(
                header=dict(values=[node],
                    align='left'),
                cells=dict(values=[[str(bottle_statuses[node])]],
                   align='left')), row_index, 1,)
            
        # pie chart for Fidelity
        data = {}
        data["values"] = [calculated_things[node]["num_correct"], calculated_things[node]["num_wrong"]]
        data["angle"] = [val / sum(data["values"]) * 2*pi for val in data["values"]]
        data['percentage'] = [val / sum(data["values"]) * 100 for val in data["values"]]
        data['percentage'] = ['{0:.2f}%'.format(val) for val in data['percentage']]
        data["class"] = ["num_correct", "num_wrong"]
        data['color'] = ["green", "red"]
        
        fig.add_trace(
        go.Pie(labels=data["class"], values=data["values"]), row_index, 2,)
    
#         fig.update_traces(hole=.4, hoverinfo="label+percent+name", selector=dict(type="domain"))

        # pie chart for class distribution
        data = {}
        freqs = dict(collections.Counter(calculated_things[node]["node_predictions"]))
        data["values"] = [val for val in freqs.values()]
        data["class"] = [val for val in freqs.keys()]
        data["angle"] = [val / sum(data["values"]) * 2*pi for val in data["values"]]
        data['percentage'] = [val / sum(data["values"]) * 100 for val in data["values"]]
        data['percentage'] = ['{0:.2f}%'.format(val) for val in data['percentage']]
        
        fig.add_trace(
        go.Pie(labels=data["class"], values=data["values"]), row_index, 3,)
        
        fig.update_traces(hole=.4, hoverinfo="label+percent+name", selector=dict(type="pie"))
        row_index += 1
        
    # make confusion matrix
    classes = data["class"]
    
    data["actual"] = [[val] * 3 for val in classes]
    data["actual"] = [item for sublist in data["actual"] for item in sublist]
    data["predicted"] = [classes] * 3
    data["predicted"] = [item for sublist in data["predicted"] for item in sublist]
        
    # go through all combinations of classes for hive since it would be the final prediction
    # ****can be modified so that each node gets their own confusion matrix****
    data["frequency"] = [calculated_things["hive"][actual][predicted] 
                       for actual, predicted in zip(data["actual"], data["predicted"])]
    data["% of results"] = [calculated_things["hive"][actual][predicted] / 
                       (calculated_things["hive"]["num_correct"] + 
                        calculated_things["hive"]["num_wrong"]) * 100 
                        for actual, predicted in zip(data["actual"], data["predicted"])]
    
    data["% of results"] = ['{0:.2f}%'.format(val) for val in data["% of results"]]
        
    fig.add_trace(go.Table(
            header=dict(values=["Actual", "Predicted", "Frequency", "% of results"],
                align='left'),
            cells=dict(values=[data["actual"], data["predicted"], data["frequency"], data["% of results"]],
                align='left'), ), row_index, 1)
    
    fig.update_layout(height=height, width=width, title_text="Classification Report")
    fig.update_traces(
        cells_font=dict(size = 15),
        selector=dict(type="table"))
    fig.show()


'''Implements the COMCOM interface.'''
import functools
import json
from typing import Dict, List, Any, Tuple, Union
import requests
'''Implements the COMCOM interface.'''


class COMCOMQueryError(BaseException):
    """Raised if any query to any node returns an error."""
    pass


class COMCOMConnectionError(BaseException):
    """Raised if connecting to any node returns an error."""
    pass


def _ensure_connected(f):
    @functools.wraps(f)
    def inner(self, *args, **kwargs):
        if not self._connected:
            raise COMCOMConnectionError(
                'Not connected to a COMCOM instance. You must call `connect()` on a COMCOMClient instance before making queries'
            )
        return f(self, *args, **kwargs)

    return inner


def _remove_unique_id(response: dict) -> dict:
    """Return *response* with the key 'unique_id' removed regardless of nesting."""
    if isinstance(response, dict):
        if 'unique_id' in response:
            del response['unique_id']
        for value in response.values():
            if isinstance(value, dict):
                _remove_unique_id(value)
    return response


class COMCOMClient:
    '''Interface for interacting with COMCOM. Creating/interacting with input_slots
       and connecting agents to COMCOM'''

    def __init__(self, comcom_info: dict, verify: bool = True):
        self.session = requests.Session()
        self._comcom_info = comcom_info
        self.name = comcom_info['name']
        self._domain = comcom_info['domain']
        self._api_key = comcom_info['api_key']
        self._headers = {'X-API-KEY': self._api_key}
        self._connected: bool = False
        self.send_unique_ids: bool = True
        self._verify: bool = verify

        if self._comcom_info.get('secure', True):
            self._secure = True
            if not self.name:
                self._url = 'https://{domain}/'.format(**self._comcom_info)
            else:
                self._url = 'https://{name}.{domain}/'.format(
                    **self._comcom_info)
        else:
            self._secure = False
            if not self.name:
                self._url = 'http://{domain}/'.format(**self._comcom_info)
            else:
                self._url = 'http://{name}.{domain}/'.format(
                    **self._comcom_info)

    def connect(self) -> Dict:
        """
        Establishes initial connection to COMCOM, then allows calling of other function to 
        """
        response_data = self.session.get(
            self._url + 'connect', verify=self._verify, headers=self._headers).json()

        if 'status' not in response_data or response_data['status'] != 'okay':
            self._connected = False
            raise COMCOMConnectionError("Connection failed!", response_data)

        if response_data['status'] == 'okay':
            self._connected = True
        else:
            self._connected = False

        return {'connection': response_data['status']}

    '''def disconnect(self) -> Dict:
        """Establishes initial connection to COMCOM, then allows calling of other function to 
        """
        response_data = self.session.get(self._url + 'connect', verify=self._verify, headers=self._headers).json()

        if 'status' not in response_data or response_data['status'] != 'okay':
            self._connected = False
            raise COMCOMConnectionError("Connection failed!", response_data)

        self._connected = False

        return {'connection': response_data['status']}'''

    @_ensure_connected
    def _query(
        self, query_method: Any, path: str, data: Union[dict, str] = None, unique_id: str = None
    ) -> Union[dict, Tuple[dict, str]]:
        """Internal helper function to make a REST API call with the given *query* and *data*."""
        if not self._connected:
            raise COMCOMConnectionError(
                'Not connected to a agent. You must call `connect()` on a AgentClient instance before making queries'
            )
        if unique_id is not None:
            if data:
                data['unique_id'] = unique_id
            else:
                data = {'unique_id': unique_id}

        data = json.loads(json.dumps(data))

        full_path = f'{self._url}{path}'
        try:
            if data is not None:
                response = query_method(full_path, verify=self._verify,
                                        headers=self._headers, json=data)
            else:
                response = query_method(
                    full_path, verify=self._verify, headers=self._headers)
            response.raise_for_status()
            response = response.json()
            if response['status'] != 'okay':
                raise COMCOMQueryError(response['message'])
            if not self.send_unique_ids:
                response = _remove_unique_id(response['message'])
            else:
                response = response['message']
        except Exception as exception:
            raise COMCOMQueryError(str(exception)) from None

        if unique_id is not None:
            return response, unique_id

        return response

    def connect_to_agent(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function to attempt to connect to an accessible agent on the network to COMCOM,
            which can then be later used for other purposes
        '''
        return self._query(self.session.post, 'connect_to_agent', data=data)

    def disconnect_agent(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call that disconnects from agent with specific agent_id, and removes it
            from all input_slots
        '''
        return self._query(self.session.post, 'disconnect_agent', data=data)

#     def modify_agent(self, data : Dict) -> Union[dict, Tuple[dict, str]]:
#         '''
#             Function call to set agent parameters such as ingress/outgress nodes, or modify genes
#         '''
#         return self._query(self.session.post, 'config_agent', data=data)

    def call_agent_command(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to pass data through comcom and call command on agent
        '''
        return self._query(self.session.post, 'call_agent_command', data=data)

    def clear_agents(self) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call that removes all connected agents in COMCOM, and in its input_slots.
            Does not delete input_slots which will continue processing data, and sending it nowhere.
        '''
        return self._query(self.session.post, 'clear_agents', data={})

    def connect_input_slot(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call which takes in config json and construction an input_slot based
            on it.
        '''
        return self._query(self.session.post, 'connect_input_slot', data=data)

    def connect_output_slot(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call which takes in config json and construction an output_slot and pipeline based
            on it.
        '''
        return self._query(self.session.post, 'connect_output_slot', data=data)

    def disconnect_input_slot(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function that takes in unique input_slot_id, and deletes it and its pipeline from COMCOM.
            Does not modify which agents are connected, and stops processing of any data sent through
            that input_slot
        '''
        return self._query(self.session.post, 'disconnect_input_slot', data=data)

    def modify_input_slot(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Complex function call which can add agents, delete agents, rearrange pipeline functions, 
            insert functon into pipeline, modify parameters, or remove functions from pipeline.
        '''
        return self._query(self.session.post, 'modify_input_slot', data=data)

    def modify_output_slot(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Complex function call which can add agents, delete agents, rearrange pipeline functions, 
            insert functon into pipeline, modify parameters, or remove functions from pipeline.
        '''
        return self._query(self.session.post, 'modify_output_slot', data=data)

    def toggle_input_slot(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function which sets flag to prevent input slot from processing and sending data to
            agents that are connected to it
        '''
        return self._query(self.session.post, 'toggle_input_slot', data=data)

    def query_input_slot(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function which sets flag to activate an agent for a single input
        '''
        return self._query(self.session.post, 'query_input_slot', data=data)

    def clear_input_slots(self) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call that tells COMCOM to clear all input slots, thus ending their data processing
        '''
        return self._query(self.session.post, 'clear_input_slots', data={})

    def load_comcom_config(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call that takes in file path, and loads in config file that comcom has access to
            might change later to simply read the json directly, but for now its a local file
        '''

        return self._query(self.session.post, 'load_comcom_config', data=data)

    def list_comcom(self) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call that takes in file path, and loads in config file that comcom has access to
            might change later to simply read the json directly, but for now its a local file
        '''

        return self._query(self.session.post, 'list_comcom', data={})

    def list_agent_connections(self) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to list all unique agent connections in comcom
        '''
        return self._query(self.session.post, 'list_agent_connections', data={})

    def list_input_slots(self) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to list all input_slots in comcom
        '''
        return self._query(self.session.post, 'list_input_slots', data={})

    def list_output_slots(self) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to list all output_slots in comcom
        '''
        return self._query(self.session.post, 'list_output_slots', data={})

    def list_preprocessing_functions(self) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to list all unique agent connections in comcom
        '''
        return self._query(self.session.post, 'list_preprocessing_functions', data={})

    def get_agent_data(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to get data about connection with passed agent_id
        '''
        return self._query(self.session.post, 'get_agent_data', data=data)

    def get_output_slot_data(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to get data about output_slot with passed output_slot_id
        '''
        return self._query(self.session.post, 'get_output_slot_data', data=data)

    def get_input_slot_data(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to get data about input_slot with passed input_slot_id
        '''
        return self._query(self.session.post, 'get_input_slot_data', data=data)

    def query_db(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to get data from db
        '''
        return self._query(self.session.post, 'query_db', data=data)

    def clear_comcom(self) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to get data from db
        '''
        return self._query(self.session.post, 'clear_comcom', data={})

    def clear_outputslot_command_queue(self, data: Dict) -> Union[dict, Tuple[dict, str]]:
        '''
            Function call to get data from db
        '''
        return self._query(self.session.post, 'clear_outputslot_command_queue', data=data)



"""Utility functions for interacting with GAIuS"""
import warnings
import json
import os
from itertools import chain
from collections import Counter
from copy import deepcopy

class GDFFormatError(BaseException):
    """Error raised when GDF is of improper format"""
    pass

def create_gdf(strings=None,
               vectors=None,
               emotives=None,
               metadata=None) -> dict:
    """Create GDF using supplied list of strings, vectors, emotives, and/or
    metadata

    Args:
        strings (list, optional): Used to provide symbols as string data
            to GAIuS. Defaults to None.
        vectors (list, optional): Used to input vector data to GAIuS.
            Defaults to None.
        emotives (dict, optional): Used to provide emotional data to GAIuS.
            Defaults to None.
        metadata (dict, optional): Used to provide miscellaneous data to GAIuS.
            Defaults to None.

    Returns:
        dict: A dictionary representing the GDF

    Example:
        .. code-block:: python

            from ia.gaius.utils import create_gdf
            gdf = create_gdf(strings=["hello"], emotives={"happy": 10.0})


    .. warning::
        If fields provided are not of the type expected, a GDFFormatError will be
        raised
    
    .. testsetup:: creategdf
        
        # here are the expected gdfs
        gdf1 = {"strings": [],
                "vectors": [],
                "emotives": {},
                "metadata": {}
                }
        gdf2 = {"strings": ["hello"],
                "vectors": [],
                "emotives": {},
                "metadata": {}
                }
        gdf3 = {"strings": ["hello"],
                "vectors": [[1, 2, 3, 4]],
                "emotives": {},
                "metadata": {}
                }
        gdf4 = {"strings": ["hello"],
                "vectors": [[1, 2, 3, 4]],
                "emotives": {"utility": 50},
                "metadata": {}
                }
        gdf5 = {"strings": ["hello"],
                "vectors": [[1, 2, 3, 4]],
                "emotives": {"utility": 50},
                "metadata": {"hello": "world"}
                }
        from ia.gaius.utils import create_gdf
        
    .. doctest:: creategdf
        :hide:
        
        >>> create_gdf() == gdf1
        True
        >>> create_gdf(strings=["hello"]) == gdf2
        True
        >>> create_gdf(strings=["hello"], vectors=[[1, 2, 3, 4]]) == gdf3
        True
        >>> create_gdf(strings=["hello"], vectors=[[1, 2, 3, 4]], emotives={"utility": 50}) == gdf4
        True
        >>> create_gdf(strings=["hello"], vectors=[[1, 2, 3, 4]], emotives={"utility": 50}, metadata={"hello": "world"}) == gdf5
        True

    """
    gdf = {
        "vectors": [] if vectors is None else vectors,
        "strings": [] if strings is None else strings,
        "emotives": {} if emotives is None else emotives,
        "metadata": {} if metadata is None else metadata
    }

    if not isinstance(gdf['vectors'], list):
        raise GDFFormatError(f"vectors field is of type \
                                  {type(gdf['vectors'])}, expected list")
    for v in gdf['vectors']:
        if not isinstance(v, list):
            raise GDFFormatError(f'Vector at index {gdf["vectors"].index(v)} is not a list')
    if not isinstance(gdf['strings'], list):
        raise GDFFormatError(f"strings field is of type \
                                  {type(gdf['strings'])}, expected list")
    if not isinstance(gdf['emotives'], dict):
        raise GDFFormatError(f"emotives field is of type \
                                  {type(gdf['emotives'])}, expected dict")
    if not isinstance(gdf['metadata'], dict):
        raise GDFFormatError(f"metadata field is of type \
                                  {type(gdf['metadata'])}, expected dict")

    return gdf


def log_progress(sequence, every=None, size=None, name='Items'): # pragma: no cover
    """
    A nice little Jupyter progress bar widget from:
    https://github.com/alexanderkuk/log-progress
    """
    from ipywidgets import IntProgress, HTML, VBox
    from IPython.display import display

    is_iterator = False
    if size is None:
        try:
            size = len(sequence)
        except TypeError:
            is_iterator = True
    if size is not None:
        if every is None:
            if size <= 200:
                every = 1
            else:
                every = int(size / 200)     # every 0.5%
    else:
        assert every is not None, 'sequence is iterator, set every'

    if is_iterator:
        progress = IntProgress(min=0, max=1, value=1)
        progress.bar_style = 'info'
    else:
        progress = IntProgress(min=0, max=size, value=0)
    label = HTML()
    box = VBox(children=[label, progress])
    display(box)

    index = 0
    try:
        for index, record in enumerate(sequence, 1):
            if index == 1 or index % every == 0:
                if is_iterator:
                    label.value = '{name}: {index} / ?'.format(
                        name=name,
                        index=index
                    )
                else:
                    progress.value = index
                    label.value = u'{name}: {index} / {size}'.format(
                        name=name,
                        index=index,
                        size=size
                    )
            yield record
    except Exception as e:
        print(f'Error in log_progress function: {str(e)})')
        progress.bar_style = 'danger'
        raise
    else:
        progress.bar_style = 'success'
        progress.value = index
        label.value = "{name}: {index}".format(
            name=name,
            index=str(index or '?')
        )


def abstract_names(ensemble: list) -> list:
    """Get a set of model names from a prediction ensemble

    Args:
        ensemble (list): a prediction ensemble

    Returns:
        list: list of models from predictions in the prediction ensemble

    Example:

        .. code-block:: python

            from ia.gaius.agent_client import AgentClient
            from ia.gaius.utils import abstract_names
            ...
            agent = AgentClient(agent_info)
            agent.connect()
            ...
            ensemble = agent.get_predictions(nodes=['P1'])
            models = abstract_names(ensemble)

    .. testsetup:: abstract_names
    
        # example prediction sequences
        ensemble1 = []
        ensemble2 = [{"name": "MODEL|1"},
                     {"name": "MODEL|2"},
                     {"name": "MODEL|3"},
                     {"name": "MODEL|4"},
                     {"name": "MODEL|5"}]
        ensemble3 = [{"name": "MODEL|0"},
                     {"name": "MODEL|0"},
                     {"name": "MODEL|0"},
                     {"name": "MODEL|0"},
                     {"name": "MODEL|0"}]
        from ia.gaius.utils import abstract_names

    .. doctest:: abstract_names
        :hide:
        
        >>> abstract_names(ensemble1) == []
        True
        >>> sorted(abstract_names(ensemble2)) == sorted(["MODEL|1", "MODEL|2", "MODEL|3", "MODEL|4", "MODEL|5"])
        True
        >>> abstract_names(ensemble3) == ['MODEL|0']
        True

    """
    return list(set([pred['name'] for pred in ensemble]))


def write_gdf_to_file(directory_name: str,
                      filename: str,
                      sequence: list) -> str:
    """Write a GDF sequence to a file

    Args:
        directory_name (str, required): directory to save GDFs to
        filename (str, required): filename to save to
        sequence (list, required): list of individual GDF events
            making up a sequence

    Example:
        .. code-block:: python

            from ia.gaius.utils import write_gdf_to_file, create_gdf
            sequence = [create_gdf(strings=["hello"]),
                        create_gdf(strings=["world"])]
            filename = 'hello_world'
            directory_name = '/example/dir'
            write_gdf_to_file(directory_name, filename, sequence)

    .. warning::
        Will overwrite the file at ``<directory_name>/<filename>``.
        Please ensure it is acceptable to do so.
        No safety checks are performed in this function

    """
    gdf_file_path = os.path.join(directory_name, filename)
    with open(gdf_file_path, 'w') as f:
        for event_idx, event in enumerate(sequence):
            json.dump(event, f)
            if event_idx != len(sequence) - 1:
                f.write('\n')

    return 'success'


def load_sequence_from_file(directory_name: str,
                      filename: str) -> list:
    """Load a GDF sequence to a file

    Args:
        directory_name (str, required): directory to load GDFs from
        filename (str, required): filename to load from

    Example:
        .. code-block:: python

            from ia.gaius.utils import load_sequence_from_file, create_gdf
            sequence = [create_gdf(strings=["hello"]),
                        create_gdf(strings=["world"])]
            filename = 'hello_world'
            directory_name = '/example/dir'
            load_sequence_from_file(directory_name, filename)


    """
    gdf_file_path = os.path.join(directory_name, filename)
    with open(gdf_file_path, 'r') as f:
        sequence = [json.loads(line) for line in f.readlines()]

    return sequence


def retrieve_bottom_level_records(traceback: dict) -> list:
    """Retrieve all records from a traceback
    (:func:`ia.gaius.agent_client.AgentClient.investigate_record`)
    call that have bottomLevel=True

    Args:
        traceback (dict): the dictionary pertaining to the output
            of an investigate call

    Returns:
        list: list of records from the traceback

    Example:
        .. code-block:: python

            from ia.gaius.agent_client import AgentClient
            from ia.gaius.utils import retrieve_bottom_level_records
            ...
            agent = AgentClient(agent_info)
            ...
            traceback_output = agent.investigate_record(record=record,
                                                        node=['P1'])
            bottom_level = retrieve_bottom_level_records(traceback_output)

    """
    bottom_level_records = []
    if traceback['bottomLevel'] is not True:
        for item_list in traceback['subitems']:
            for item in item_list:
                if isinstance(item, dict):
                    bottom_level_records.extend(retrieve_bottom_level_records(deepcopy(item)))
    else:
        bottom_level_records.append(traceback)

    return bottom_level_records

def merge_gdfs(gdf1: dict, gdf2: dict) -> dict:
    """Merge two GDFs into a single gdf, accumulating the values in each field

    Args:
        gdf1 (dict): First GDF
        gdf2 (dict): Second GDF

    Raises:
        Exception: When vectors are of differing lengths

    Returns:
        dict: Merged GDF
    """

    merge_strings = list(chain(gdf1["strings"], gdf2["strings"]))

    merge_vecs = list(chain(gdf1["vectors"], gdf2["vectors"]))
    print(f"{merge_vecs=}")
    if len(merge_vecs) > 0:
        if not all([len(vec) == len(merge_vecs[0]) for vec in merge_vecs]):
            raise Exception(f"Vectors not all of same length!!!")

    merge_emotives = Counter()
    merge_emotives.update(gdf1["emotives"])
    merge_emotives.update(gdf2["emotives"])
    merge_emotives = dict(merge_emotives)

    # no way to get around conflicts here, just going to add keys from gdf1, then update with keys from gdf2
    merge_metadata = dict()
    merge_metadata.update(gdf1.get("metadata", {}))
    merge_metadata.update(gdf2.get("metadata", {}))

    return create_gdf(strings=merge_strings, vectors=merge_vecs, emotives=merge_emotives, metadata=merge_metadata)



import plotly.express as px
import plotly.graph_objects as go
import json
import pandas as pd
import numpy as np
from ia.gaius.agent_client import AgentClient
from copy import deepcopy


def make_classification_pie_chart(classification_counts):
    """Return a pie chart of classification frequency distribution"""
    class_df = pd.DataFrame([dict(Classification=k, Count=v) for k, v in classification_counts.items()])
    class_df.sort_values(by=['Classification'], inplace=True)
    classification_graph = px.pie(class_df, values='Count', names="Classification", title="Classification Distribution")
    return classification_graph


class offline_sio:
    def __init__(self, agent: AgentClient = None):
        print('init offline sio')
        self.log = []
        self.agent = agent

    def emit(self, topic, message, to=None):
        # print(f'broadcasting on topic {topic} to {to}: {message}')
        self.log.append(deepcopy(message))

    def dump(self, filename):
        with open(filename, 'w') as f:
            json.dump(self.log, f)

    def load(self, filename):
        with open(filename, 'r') as f:
            self.log = json.load(f)

    def clear(self):
        self.log = []

    def make_report(self, filepath="report.html"):
        """Generate plotly report from PVT results captured in self.log

        Args:
            filepath (str, optional): Path at which to save the html report. Defaults to "report.html".

        Returns:
            str: path of html report
        """
        final_record = [x for x in self.log if x['status'] == 'finished']
        assert (len(final_record) == 1)

        # data from final entry
        final_record = final_record[0]

        print(f'{final_record.keys()}')
        print(f'{final_record["metrics"].keys()}')
        classification_counts = final_record['metrics']['classification_counter']
        cg = make_classification_pie_chart(classification_counts)

        testing_records = [deepcopy(x) for x in self.log if x['status'] == 'testing']
        print(f'{len(testing_records)=}')

        testing_records = sorted(testing_records, key=lambda x: x['current_record'])
        testing_record_count = len(testing_records)
        if self.agent:
            for tmp_record in testing_records:
                tmp_record['metrics']['recall_threshold'] = self.agent.get_gene('recall_threshold')

        nodes_df = []
        print(f"{testing_records[0]['metrics']['predicted'].keys()=}")

        node_data = self.agent.get_all_genes()
        flattened_node_data = []
        for k, v in node_data.items():
            new_v = {**v['genes']}
            new_v['node_name'] = k
            new_v['node_id'] = self.agent.genome.primitive_map[k]
            flattened_node_data.append(deepcopy(new_v))

        flattened_node_data = sorted(flattened_node_data, key=lambda x: x['node_name'])
        node_data = pd.DataFrame(flattened_node_data)
        # display(node_data)
        node_fig = go.Figure(data=[go.Table(header=dict(values=list(node_data.columns),
                                                        fill_color='paleturquoise',
                                                        align='left'),
                                            cells=dict(values=node_data.transpose().values.tolist(),
                                                       fill_color='lavender',
                                                       align='left'))])
        node_fig.show()

        for node in testing_records[0]['metrics']['predicted'].keys():
            for record in testing_records:
                tmp_record = deepcopy(record)

                del tmp_record['metrics']['classification_counter']
                del tmp_record['metrics']['idx']
                actual = tmp_record['metrics']['actual']
                del tmp_record['metrics']['actual']

                # Update record with node specific info
                # print(f'{tmp_record["metrics"]=}')
                tmp_record['metrics']['recall_threshold']['hive'] = 1.0
                tmp_record.update({k: v[node] for k, v in tmp_record['metrics'].items()})
                tmp_record['node'] = node
                tmp_record['actual'] = actual
                if 'recall_threshold' in tmp_record:
                    if isinstance(tmp_record['recall_threshold'], dict):
                        tmp_record['recall_threshold'] = tmp_record['recall_threshold']['recall_threshold']
                nodes_df.append(tmp_record)

        # make dataframe out of node specific testing records
        nodes_df = pd.DataFrame(nodes_df)
        accuracies_g = px.line(nodes_df,
                               x='current_record',
                               y='accuracy',
                               range_y=[0, 1],
                               title="Running Accuracy per Node",
                               color='node',
                               markers=False,
                               labels={"current_record": "Testing Record index",
                                       "accuracy": "Running Accuracy (0.0 to 1.0)",
                                       "node": "Cognitive Processor"
                                       }
                               )
        precisions_g = px.line(nodes_df,
                               x='current_record',
                               y='precision',
                               range_y=[0, 1],
                               title="Running Precision per Node",
                               color='node',
                               labels={"current_record": "Testing Record index",
                                       "precision": "Running Precision (0.0 to 1.0)",
                                       "node": "Cognitive Processor"
                                       }
                               )
        resp_perct_g = px.line(nodes_df,
                               x='current_record',
                               y='response_percentage',
                               range_y=[0, 1],
                               title="Response Percentage per Node",
                               color='node',
                               labels={"current_record": "Testing Record index",
                                       "response_percentage": "Response Percentage (0.0 to 1.0)",
                                       "node": "Cognitive Processor"
                                       }
                               )

        # fig2 = go.Figure(data=[go.Surface(y=nodes_df['recall_threshold'], x=nodes_df['current_record'], z=nodes_df['precision'])])
        # fig2.show()
        fig = px.scatter_3d(nodes_df, y='recall_threshold', x='current_record', z='accuracy', color='node')
        fig.update_traces(marker_size=1)

        fig2 = px.line_3d(nodes_df, y='recall_threshold', x='current_record', z='precision', line_group='node', color='node')
        fig2.update_traces(marker_size=1)

        # x = nodes_df['current_record']
        # y = nodes_df['recall_threshold']
        z = nodes_df['precision']

        z2 = nodes_df[['precision']].copy()
        z2 = np.tile(z2, [z.shape[0], 1])

        fig3 = None

        # split testing records into 100 parts
        # get round increment to nearest 10
        fig3_increment = round(testing_record_count / 100)
        fig3_increment = round(fig3_increment / 10) * 10
        # print(f'{fig3_increment=}')

        if fig3_increment >= 10:
            fig3 = go.Figure()
            for step in range(0, testing_record_count, fig3_increment):
                subset = nodes_df.loc[nodes_df['current_record'] == step]
                fig3.add_trace(go.Scatter(visible=False,
                                          line=dict(color="#00CED1", width=2),
                                          name=f'current_record={step}',
                                          x=subset['recall_threshold'],
                                          y=subset['precision'],
                                          mode='lines+markers',
                                          hoverinfo='text',
                                          text="Node: " + subset['node'].astype(str) + "<br>Running Precision: " + subset['precision'].round(4).astype(str) + "<br> Running Accuracy: " + subset['accuracy'].round(4).astype(str),
                                          )
                               )
            fig3.data[0].visible = True

            steps = []
            for i in range(len(fig3.data)):
                step = dict(
                    method="update",
                    args=[{"visible": [False] * len(fig3.data)},
                          {"title": f"Recall threshold vs Running Precision at record {str(i * fig3_increment)}"}],  # layout attribute
                    label=str(i * fig3_increment)
                )
                step["args"][0]["visible"][i] = True  # Toggle i'th trace to "visible"
                steps.append(step)

            sliders = [dict(
                active=0,
                currentvalue={"prefix": "Current Record: "},
                pad={"t": 50},
                steps=steps
            )]

            fig3.update_layout(
                sliders=sliders,
                title=dict(text="Recall threshold vs Running Precision at record 0"),
                xaxis=dict(title="Recall Threshold"),
                yaxis=dict(title="Running Precision"),
            )
            fig3.show()

        # display everything
        fig.show()
        fig2.show()
        accuracies_g.show()
        precisions_g.show()
        resp_perct_g.show()
        cg.show()

        with open(filepath, 'w') as f:
            f.write(node_fig.to_html(include_plotlyjs='True', full_html=False, default_width='80%', default_height='50%'))
            f.write(accuracies_g.to_html(include_plotlyjs=False, full_html=False, default_width='80%', default_height='50%'))
            f.write(precisions_g.to_html(include_plotlyjs=False, full_html=False, default_width='80%', default_height='50%'))
            f.write(resp_perct_g.to_html(include_plotlyjs=False, full_html=False, default_width='80%', default_height='50%'))
            f.write(cg.to_html(include_plotlyjs=False, full_html=False, default_width='80%', default_height='50%'))
            f.write(fig.to_html(include_plotlyjs=False, full_html=False, default_width='80%', default_height='50%'))
            f.write(fig2.to_html(include_plotlyjs=False, full_html=False, default_width='80%', default_height='50%'))
            if fig3:
                f.write(fig3.to_html(include_plotlyjs=False, full_html=False, default_width='80%', default_height='50%'))
        return filepath



import os
import random
import re
from pathlib import Path
from typing import Union


def validate_data(data: dict):
    """Validates if the data is in correct GAIuS digestible format.
    Returns True if data validates.  Returns False if data does not validate.

    Args:
        data (dict, required): GDF to validate

    Example:
        .. code-block:: python

            >>> gdf = {'strings':["hello"], 'vectors': [], 'emotives': {} }
            >>> validate_data(gdf)
            True
            >>> bad_gdf = {'strings': []}
            >>> validate_data(bad_gdf)
            Exception: Dictionary requires "vectors", "emotives", and "strings" as keys!
            >>> bad_gdf_2 = {'strings':["hello"], 'vectors': [], 'emotives': ['utility|5'] }
            >>> validate_data(bad_gdf_2)
            Exception: "emotives" must be a dict of <str, float>. Dict not provided!

    .. testsetup:: validate_data

        gdf1 = {"strings": [],
            "vectors": [],
            "emotives": {},
            "metadata": {}}

        # empty gdf with additional field
        gdf2 = {"strings": [],
                "vectors": [],
                "emotives": {},
                "metadata": {},
                "invalid": True}

        # gdf with single string
        gdf3 = {"strings": ["hello"],
                "vectors": [],
                "emotives": {},
                "metadata": {}
                }

        # gdf with single string and vector
        gdf4 = {"strings": ["hello"],
                "vectors": [[1, 2, 3, 4]],
                "emotives": {},
                "metadata": {}
                }

        # gdf with malformed strings field
        gdf5 = {"strings": "hello",
                "vectors": [],
                "emotives": {},
                "metadata": {}
                }

        # gdf with malformed vectors field
        gdf6 = {"strings": ["hello"],
                "vectors": [1, 2, 3, 4],
                "emotives": {},
                "metadata": {}
                }

        # gdf with malformed emotives field
        gdf7 = {"strings": ["hello"],
                "vectors": [[1, 2, 3, 4]],
                "emotives": {"utility": "high"},
                "metadata": {}
                }

        # gdf with valid emotives field
        gdf8 = {"strings": ["hello"],
                "vectors": [[1, 2, 3, 4]],
                "emotives": {"utility": 23.7},
                "metadata": {}
                }
                
        from ia.gaius.data_ops import validate_data

    .. doctest:: validate_data
        :hide:
        
        >>> validate_data(gdf1)
        True
        >>> validate_data(gdf2)
        Traceback (most recent call last):
        ...
        Exception: Key: %s, should not be in the data dictionary!
        >>> validate_data(gdf3)
        True
        >>> validate_data(gdf4)
        True
        >>> validate_data(gdf5)
        Traceback (most recent call last):
        ...
        Exception: "strings" must be a list of strings.  List not provided!
        >>> validate_data(gdf6)
        Traceback (most recent call last):
        ...
        Exception: "vectors" must be a list of arrays (i.e. lists)!
        >>> validate_data(gdf7)
        Traceback (most recent call last):
        ...
        Exception: "emotives" must be a dict of <str, float>!
        >>> validate_data(gdf8)
        True
        
    """
    if not isinstance(data, dict):
        raise Exception("Incorrect data type.  Must be a dictionary.")
    if "vectors" not in list(data.keys()) or "emotives" not in list(data.keys()) or "strings" not in list(
            data.keys()):
        raise Exception('Dictionary requires "vectors", "emotives", and "strings" as keys!')
    for key in list(data.keys()):
        if key not in ["vectors", "emotives", "strings", "metadata"]:
            raise Exception("Key: %s, should not be in the data dictionary!")

    if not isinstance(data["strings"], list):
        raise Exception('"strings" must be a list of strings.  List not provided!')
    for item in data["strings"]:
        if not (isinstance(item, str) or isinstance(item, str)):
            raise Exception('"strings" must be a list of strings or unicode objects!')

    if not isinstance(data["emotives"], dict):
        raise Exception('"emotives" must be a dict of <str, float>. Dict not provided!')
    for item in data["emotives"].values():
        if not isinstance(item, (int, float)):
            raise Exception('"emotives" must be a dict of <str, float>!')

    if not isinstance(data["vectors"], list):
        raise Exception('"vectors" must be a list of arrays.  List not provided!')
    for item in data["vectors"]:
        if not isinstance(item, list):
            raise Exception('"vectors" must be a list of arrays (i.e. lists)!')
    return True


def raw_in_count(filename: str):
    return int(os.popen("sed -n '$=' '%s'" % filename).readline().split()[0])


def atoi(text: str):
    """Attempt to convert string to int"""
    return int(text) if text.isdigit() else text


def natural_keys(text: str):
    """
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    """
    return [atoi(c) for c in re.split(r'(\d+)', str(text))]


class DataRecords:
    """Splits data into random sets for training and testing."""

    def __init__(self, original_dataset: Union[str, Path], DR: float, DF: float, shuffle: bool, folder: bool = True):
        """
        Args:
            original_dataset (str or list, required): location of dataset to use for training and testing sets
            DR (float, required): fraction of total data to use for testing and training. 0 < DR < 100
            DF (float, required): fraction of the DR to use for training.  The rest of the DR is used for testing. 0 < DF < 100
            shuffle (bool, required): whether to shuffle the data when creating sets
            folder (bool, optional): set if the original dataset is a folder

        After creating the class, utilize the member variables `train_sequences` and `test_sequences` for the data sets

        :ivar train_sequences: the files to use for training
        :ivar test_sequences: the files to use for testing
        """
        
        if folder:
            original_dataset = [Path(original_dataset).joinpath(f) for f in os.listdir(original_dataset) if
                                not f.startswith('.') and not f.startswith('_')]
            original_dataset.sort(key=natural_keys)
        DR = DR / 100
        DF = DF / 100
        try:
            if DR == 1 and DF == 1:
                if shuffle:
                    random.shuffle(original_dataset)
                self.train_sequences = original_dataset
                self.test_sequences = []
            else:
                if shuffle:
                    random.shuffle(original_dataset)

                num_files = len(original_dataset)
                num_use_files = int(num_files * DR)  # use a fraction of the whole set

                num_train_sequences = int(num_use_files * DF)  # train 2/3rds, test 1/3rd
                num_test_sequences = num_use_files - num_train_sequences

                self.train_sequences = original_dataset[:num_train_sequences]
                self.test_sequences = original_dataset[num_train_sequences:(num_train_sequences + num_test_sequences)]
        except Exception as exception:
            print(f'DataRecords BROKE by {exception.args}')
            raise


class Data:
    def __init__(self, data_directories=None, dataset=None):
        """Supply either a list of data_directories, or a dataset."""
        self.dataset = None
        self.data_directories = None
        if data_directories is not None:
            self.data_directories = data_directories
        elif dataset is not None:
            self.dataset = dataset
        self.train_sequences = []
        self.test_sequences = []
        self.percent_of_dataset_chosen = None
        self.percent_reserved_for_training = None

    def prep(self, percent_of_dataset_chosen: float, percent_reserved_for_training: float, shuffle: bool = False):
        
        self.percent_of_dataset_chosen = percent_of_dataset_chosen
        self.percent_reserved_for_training = percent_reserved_for_training
        
        if self.data_directories:
            data = [DataRecords(Path(d), percent_of_dataset_chosen, percent_reserved_for_training, shuffle, folder=True) for d
                    in self.data_directories]  # It's a list because the user may pick several data file directories.
        elif self.dataset:
            data = [DataRecords(self.dataset, percent_of_dataset_chosen, percent_reserved_for_training, shuffle,
                                folder=False)]

        self.train_sequences = []
        self.test_sequences = []
        for d in data:
            self.train_sequences += d.train_sequences
            self.test_sequences += d.test_sequences

class PreparedData(Data):
    """Overloaded type for Data class to signify that train_sequences and test_sequences
    contain raw sequences, and not filepaths to sequences
    """
    def __init__(self, data_directories=None, dataset=None):
        super().__init__(data_directories, dataset)
        pass


import docker
import docker.errors as de
import logging


logging.basicConfig(format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
                    datefmt='%Y-%m-%d:%H:%M:%S')
logger = logging.getLogger(__name__)


def spawn_container(container_name: str,
                    api_key: str,
                    ports: dict,
                    docker_image: str,
                    privileged: bool = False,
                    volumes: dict = None,
                    env: dict = None,
                    hostname: str = None):

    client = docker.from_env()

    if env is None:
        env = {}

    env['API_KEY'] = api_key

    if volumes is None:
        volumes = {}

    if hostname is None:
        hostname = container_name
    
    try:
        client.containers.run(image=docker_image,
                              init=True,
                              ports=ports,
                              privileged=privileged,
                              name=container_name,
                              volumes=volumes,
                              environment=env,
                              detach=True,
                              hostname=hostname)

    except de.ImageNotFound as e:
        logger.error(f'Image not found: {str(e)}')
        raise Exception(f'Failed to start Container: {str(e)}')

    except de.APIError as e:
        logger.error(f'Docker API Error: {str(e)}')
        raise Exception(f'Failed to start Container: {str(e)}')

    return True

def create_docker_network(network_name: str) -> None:
    
    try:
        # create docker network for comcom
        client = docker.from_env()
        client.networks.create(name=network_name)
    
    except de.APIError as docker_error:
        logger.exception('Exception while making docker network')
        raise

def kill_container(container_name: str) -> bool:

    client = docker.from_env()

    try:
        logger.debug(f"attempting to remove container {container_name}")
        tf = client.containers.get(container_name)
        # tf.kill()
        tf.remove(force=True)
    except de.NotFound as e:
        logger.error(
            f'Container {container_name} not found: {str(e)}')
        return True
    except de.APIError as e:
        logger.error(f'Docker API Error: {str(e)}')
        raise Exception(f'Failed to kill container {str(e)}')
    return True



def connect_networks_container(networks: list, container_name: str):
    client = docker.from_env()

    try:
        container = client.containers.get(container_name)

        docker_networks = client.networks.list(names=networks)
        for net in docker_networks:
            logger.info(f'Connecting to {net.name}')
            try:
                net.connect(container)
            except de.APIError as e:
                logger.error(f'Docker API Error: {str(e)}')
                pass  # already connected on network
    except de.NotFound as e:
        logger.error(
            f'Container {container_name} not found: {str(e)}')
        return False
    except de.APIError as e:
        logger.error(f'Docker API Error: {str(e)}')
        raise Exception(f'Docker API Error: {str(e)}')

    return True


def get_docker_bridge_ip() -> str:
    client = docker.from_env()

    try:

        bridge = client.networks.get('bridge')
        bridge_ip: str = bridge.attrs['IPAM']['Config'][0]['Gateway']
        logger.debug(f'Docker bridge IP = {bridge_ip}')
        return bridge_ip

    except Exception as bridge_error:
        logger.exception(
            f'Exception while retrieving docker bridge IP address: {str(bridge_error)}')
        raise



import random
import pymongo
import json
from bson.objectid import ObjectId
from copy import deepcopy
import zipfile
import datetime
from traceback import print_exc
import io
from ia.gaius.data_ops import validate_data


class MongoDataRecords:
    def __init__(self,
                 dataset_records,
                 DR: float,
                 DF: float,
                 shuffle: bool):
        """
        Args:
            dataset_records (str or list, required): List of mongo
                ObjectIds to use as data records
            DR (float, required): fraction of total data to use for
                testing and training. 0 < DR < 100
            DF (float, required): fraction of the DR to use for training.
                The rest of the DR is used for testing. 0 < DF < 100
            shuffle (bool, required): whether to shuffle the
                data when creating sets

            After creating the class, utilize the member variables
                `train_sequences` and `test_sequences` for the data sets

        :ivar train_sequences: the mongo documents to use for training
        :ivar test_sequences: the mongo documents to use for testing
        """
        DR = DR / 100
        DF = DF / 100
        self.random = random.Random()

        if DR > 1 or DF > 1:
            raise Exception(f'Invalid value provided for DR or DF: \
                            {DR=}, {DF=}')

        try:
            if shuffle:
                self.random.shuffle(dataset_records)
            if DR == 1 and DF == 1:
                self.train_sequences = dataset_records
                self.test_sequences = []
            else:
                num_files = len(dataset_records)
                # use a fraction of the whole set
                num_use_files = int(num_files * DR)

                num_train_sequences = int(num_use_files * DF)
                num_test_sequences = num_use_files - num_train_sequences

                self.train_sequences = dataset_records[:num_train_sequences]
                self.test_sequences = dataset_records[num_train_sequences:(
                    num_train_sequences + num_test_sequences)]
        except Exception as exception:
            print(f'MongoDataRecords BROKE by {exception.args}')
            raise


def check_dataset_exists(mongo_db,
                         dataset_collection_name: str,
                         user_id: str,
                         dataset_id: str):
    """Will check if the specified dataset is found in MongoDB

    Args:
        mongo_db : MongoDB database object
        dataset_collection_name (str): collection name where
            master dataset record is saved
        user_id (string): user unique identifier
        dataset_id (string): dataset unique identifier

    Returns:
        bool: represents whether the dataset was successfully found
    """
    dataset_obj = {'user_id': user_id,
                   'dataset_id': dataset_id}

    # checking if dataset count != 1. If >1, something nefarious is
    # potentially happening
    if mongo_db[dataset_collection_name].count_documents({**dataset_obj}) != 1:
        return False
    return True


class MongoData:
    """
        Analogous object to the Data class, but utilizes a MongoDB cursor
        instead of a directory to reference data records

        Start with a MongoDB document containing the name of all dataset
        files (located separately)

        Only retrieve actual data files when calling retrieveDataRecord.
        Overloaded iterator functions to allow treating of object
        as a list.

        Example:

            .. code-block:: python

                >>> mongo = pymongo.MongoClient('mongodb://mongodb:27017/')
                >>> mongo_db = mongo.db['main_database']
                >>> dataset_details = {"user_id": "ABCD1",
                                    "dataset_id": "iris_0_0_13"}
                >>> md = MongoData(mongo_dataset_details=dataset_details,
                                mongo_db=mongo_db,
                                data_files_collection_name='dataset_files')
                >>> md.prep(percent_of_dataset_chosen=50,
                            percent_reserved_for_training=50,
                            shuffle=True)
                >>> md.setIterMode('testing')
                >>> for record in md:
                ...

    """

    def __init__(self,
                 mongo_dataset_details: dict,
                 data_files_collection_name: str,
                 mongo_db: pymongo.MongoClient,
                 dataset_collection_name: str = 'datasets'):
        """Initialized dataset object from MongoDB

        Args:
            mongo_dataset_details (dict): contains info about the user_id
                and dataset_id of the dataset. Used to query MongoDB
            data_files_collection_name (str): Collection in MongoDB where
                individual data records are stored
            dataset_collection_name (str): Collection in MongoDB where master
                dataset info record is stored
            mongo_db (pymongo.MongoClient): MongoDB client object to use
                for dataset lookups

        Raises:
            Exception: user_id field missing from mongo_dataset_details
            Exception: dataset_id field missing from mongo_dataset_details
            Exception: multiple datasets found pertaining to same user_id
                and dataset_id field
        """
        if 'user_id' not in mongo_dataset_details:
            raise Exception(f'user_id field missing from mongo_dataset_details\
                            ({mongo_dataset_details=})')
        if 'dataset_id' not in mongo_dataset_details:
            raise Exception(f'dataset_id field missing from \
                            mongo_dataset_details ({mongo_dataset_details=})')

        self._mongo_dataset_details = deepcopy(mongo_dataset_details)
        self._data_files_collection_name = deepcopy(data_files_collection_name)
        self.mongo_db = mongo_db

        # initialize iteration variables
        self._iter_mode = None
        self.__curr = None
        self.__term = None

        self.percent_of_dataset_chosen = None
        self.percent_reserved_for_training = None
        dataset_obj = {'user_id': self._mongo_dataset_details['user_id'],
                       'dataset_id': self._mongo_dataset_details['dataset_id']}

        ds_count: int = mongo_db[dataset_collection_name].count_documents(
            dataset_obj)

        if ds_count != 1:

            raise Exception(f'{ds_count} \
                            dataset records found for {dataset_obj=}')

        dataset = mongo_db[dataset_collection_name].find_one(dataset_obj)

        self._mongo_dataset_details['files'] = \
            [item for item in dataset['dataset']['files']]

        self.train_sequences = []
        self.test_sequences = []

    @classmethod
    def upload_dataset(cls, mongo_db, dataset_details: dict, filepath: str):
        """Upload a dataset to MongoDB from a local filepath

        Args:
            mongo_db (pymongo.MongoClient): MongoDB Database object
            dataset_details (dict): Dictionary containing details about
            dataset (e.g. name, user_id, dataset_id, collection names, etc.)
            filepath (str): filepath of zip folder
                containing dataset (GDF records)

        Returns:
            _type_: _description_

        Example:
            .. code-block:: python

                from ia.gaius.pvt.mongo_interface import MongoData
                ...
                dataset_details = {"user_id": "user-1234",
                                   "dataset_name": "MNIST",
                                   "dataset_id": "abba12",
                                   "data_files_collection_name": "dataset_files",
                                   "dataset_collection_name": "datasets"}
                MongoData.upload_dataset(mongo_db=mongo_db,
                                         dataset_details=dataset_details)

        """

        data_files_collection_name = dataset_details['data_files_collection_name']
        dataset_collection_name = dataset_details['dataset_collection_name']

        # ensure dataset details has user_id and dataset_id fields
        if 'user_id' not in dataset_details:
            raise Exception(
                f'user_id field missing from dataset_details ({dataset_details=})')
        if 'dataset_id' not in dataset_details:
            raise Exception(
                f'dataset_id field missing from dataset_details ({dataset_details=})')

        cur_time = datetime.datetime.now().isoformat()
        dataset_obj = {'user_id': dataset_details['user_id'],
                       'dataset_id': dataset_details['dataset_id'],
                       'lastModifiedDate': cur_time,
                       'uploadedDate': cur_time,
                       'size': None,
                       'dataset': {'files': []}}

        # retrieve dataset from filepath
        # if anything fails along the way, try to remediate database
        # by deleting stuff already inserted
        try:
            with zipfile.ZipFile(filepath, 'r') as zipObject:
                dataset_obj['size'] = sum(
                    [zinfo.file_size for zinfo in zipObject.filelist])
                records = zipObject.namelist()
                records = [(zipObject.read(item), item)
                           for item in records if not item.endswith('/') or item.startswith('.') or item.startswith('__')]
                records = [{'file': item[0],
                            'filename': item[1],
                            'user_id': dataset_obj['user_id'],
                            'dataset_id': dataset_obj['dataset_id']}
                           for item in records]

                # upload individual records to MongoDB data_files collection
                try:
                    insert_result = mongo_db[data_files_collection_name].insert_many(
                        records)
                except Exception as e:
                    print(f'failed to insert records: {str(e)}')
                    try:
                        mongo_db[data_files_collection_name].delete_many(
                            {'dataset_id': dataset_obj['dataset_id']})
                    except Exception as e2:
                        print(
                            f'failed to remediate data files collection: {str(e2)}')
                    finally:
                        return f'failed to insert records: {str(e)}'
                dataset_items = insert_result.inserted_ids
                dataset_obj['dataset']['files'] = dataset_items
                dataset_obj['dataset']['file_count'] = len(dataset_items)

                try:
                    # upload master dataset record to MongoDB dataset_collection_name
                    mongo_db[dataset_collection_name].insert_one(dataset_obj)
                except Exception as e:
                    print(f'failed to insert master record: {str(e)}')
                    try:
                        mongo_db[dataset_collection_name].delete_one(
                            dataset_obj)
                        mongo_db[data_files_collection_name].delete_many(
                            {'dataset_id': dataset_obj['dataset_id']})
                    except Exception as e2:
                        print(f'failed to remediate database: {str(e2)}')
                    finally:
                        return f'failed to insert master dataset record: {str(e)}'

        except Exception as e3:
            print(f'failed in zipfile handling: {str(e3)}')
            print_exc()
            return f'failed in zipfile handling: {str(e3)}'

        return 'success!!!'

    @classmethod
    def delete_dataset(cls, mongo_db, dataset_details: dict):
        """Upload a dataset to MongoDB from a local filepath

        Args:
            mongo_db (pymongo.MongoClient): MongoDB database object
            dataset_details (dict): Dictionary containing details about
                dataset (e.g. name, user_id, dataset_id, collection names)

        Returns:
            str: String depicting action that was taken

        Example:
            .. code-block:: python

                from ia.gaius.pvt.mongo_interface import MongoData
                ...
                dataset_details = {"user_id": "user-1234",
                                   "dataset_name": "MNIST",
                                    "dataset_id": "abba12",
                                    "data_files_collection_name": "dataset_files",
                                    "dataset_collection_name": "datasets"}
                MongoData.delete_dataset(mongo_db=mongo_db,
                                         dataset_details=dataset_details)

        """
        data_files_collection_name = dataset_details['data_files_collection_name']
        dataset_collection_name = dataset_details['dataset_collection_name']
        try:
            if not check_dataset_exists(mongo_db=mongo_db,
                                        dataset_collection_name=dataset_collection_name,
                                        user_id=dataset_details['user_id'],
                                        dataset_id=dataset_details['dataset_id']):
                return 'dataset-not-found'
        except Exception as e:
            print(f"error while checking if dataset exists: {str(e)}")
            raise e
        dataset_obj = {'user_id': dataset_details['user_id'],
                       'dataset_id': dataset_details['dataset_id']}
        dataset = None
        try:
            if mongo_db[dataset_collection_name].count_documents({**dataset_obj}) != 1:
                raise Exception(
                    f'found {mongo_db[dataset_collection_name].count_documents({**dataset_obj})} datasets in mongo, expecting 1')
            dataset = mongo_db[dataset_collection_name].find_one(
                {**dataset_obj})
        except Exception as e:
            print(f"error while retrieving dataset: {e}")
            raise e

        try:
            mongo_db[data_files_collection_name].delete_many(
                {'_id': {'$in': dataset['dataset']['files']}})
        except Exception as e:
            print(e)
            raise e

        mongo_db[dataset_collection_name].delete_one({**dataset_obj})

        return 'deleted'

    def prep(self,
             percent_of_dataset_chosen: float,
             percent_reserved_for_training: float,
             shuffle: bool = False):
        """Prepare the dataset

        Args:
            percent_of_dataset_chosen (float): The percent of the dataset
                to utilize
            percent_reserved_for_training (float): The training/testing
                split for the dataset (e.g.
                set to 80 for 80/20 training/testing split)
            shuffle (bool, optional): Whether to shuffle the data.
                Defaults to False.
        """
        self.percent_of_dataset_chosen = percent_of_dataset_chosen
        self.percent_reserved_for_training = percent_reserved_for_training
        
        data = [MongoDataRecords(deepcopy(self._mongo_dataset_details['files']),
                                 percent_of_dataset_chosen,
                                 percent_reserved_for_training,
                                 shuffle)]

        self.train_sequences = []
        self.test_sequences = []
        for d in data:
            self.train_sequences += d.train_sequences
            self.test_sequences += d.test_sequences

    def retrieveDataRecord(self, document_id: ObjectId):
        """Retrieve a data record from MongoDB,
        pertaining to the ObjectId specifed

        Args:
            document_id (ObjectId, required): data record to retrieve
                from mongo, located in the collection specified when
                calling :func:`__init__`

        Raises:
            Exception: Raised when MongoDB document is not found.
                Shows query performed that failed

        Returns:
            str: binary string depicting data sequence
            stored in MongoDB Document
        """
        query_dict = {'user_id': self._mongo_dataset_details['user_id'],
                      'dataset_id': self._mongo_dataset_details['dataset_id'],
                      '_id': document_id}

        record = self.mongo_db[self._data_files_collection_name].find_one(
            query_dict)
        if record is None:
            raise Exception(
                f'DataRecord {str(document_id)} not found using {query_dict=}')

        return record['file']

    def convertBinaryStringtoSequence(self, record):
        """Convert Binary string of multiple GDFs (delimited by newline) into a sequence of JSON objects

        Args:
            record (str, required): binary string of GDFs to convert

        Returns:
            list: list of GDFs in json format
        """
        try:
            lines = io.StringIO(record.decode('utf-8'))
            sequence = (tuple([json.loads(line.strip()) for line in lines]))
            return sequence
        except Exception as e:
            print(f'failed to retrieve record as StringIO: {e}')

    def getSequence(self, record):
        """Wrapper function to retrieve a record from MongoDB
        and convert it into a sequence

        Args:
            record (ObjectId): The MongoDB ObjectId of the data
                record to retrieve

        Returns:
            list: GDF sequence retrieved from MongoDB
        """
        sequence = self.convertBinaryStringtoSequence(
            self.retrieveDataRecord(record))
        for gdf in sequence:
            validate_data(gdf)
        return sequence

    def setIterMode(self, mode: str) -> None:
        """Set mode to be used for iterating across dataset

        Args:
            mode (str): set to "training" or "testing" depending
                on what set of sequences is to be iterated across

        Raises:
            Exception: When no data is in train_sequences or
                test_sequences, and :func:`prep` should be called first
            Exception: When invalid mode specified in mode argument
        """
        if len(self.train_sequences) == 0 and len(self.test_sequences) == 0:
            raise Exception('no data in train_sequences or test_sequences, \
                            use prep first')

        if mode == 'training':

            self.__curr = 0
            self.__term = len(self.train_sequences)

        elif mode == 'testing':
            self.__curr = 0
            self.__term = len(self.test_sequences)
        else:
            raise Exception('setIterMode only supports "training" \
                            and "testing"')

        self._iter_mode = mode
        return

    def __iter__(self):
        return self

    def __next__(self):
        """Iterate across range, calling :func:`retrieveDataRecord` for each item
        """
        if self.__curr >= self.__term:
            self.__curr = 0
            raise StopIteration()

        if self._iter_mode == 'training':
            (cur, self.__curr) = (
                self.train_sequences[self.__curr], self.__curr + 1)

        elif self._iter_mode == 'testing':
            (cur, self.__curr) = (
                self.test_sequences[self.__curr], self.__curr + 1)

        else:
            raise Exception('invalid iter mode: {self._iter_mode}, \
                            set using setIterMode(mode)')

        return self.retrieveDataRecord(cur)


class MongoResults:
    """Class to handle saving and linking result data inside MongoDB.

    Provides functions to insert single log record during training/testing,
    save final result after test completion, and remediation/deletion
    function for test aborting, database cleanup
    """

    def __init__(self,
                 mongo_db,
                 result_collection_name: str,
                 log_collection_name: str,
                 test_id: str,
                 user_id: str,
                 dataset_id: str,
                 test_configuration: dict = None):
        """Initialize MongoResults object

        Args:
            mongo_db (pymongo.MongoClient): Database where the results
                are to be stored
            result_collection_name (str): collection name to save final
                test results
            log_collection_name (str): collection name to save testing
                log documents
            test_id (str): unique-id for the test being conducted
            user_id (str): unique-id for the user conducting the test
            dataset_id (str): unique-id for the dataset being used
                in the test
            test_configuration (dict): object showing all of the options
                used for configuring pvt
        """

        self.test_id = test_id
        self.user_id = user_id
        self.dataset_id = dataset_id

        self.mongo_db = mongo_db
        self.result_collection_name = result_collection_name
        self.log_collection_name = log_collection_name
        self.result_collection = mongo_db[result_collection_name]
        self.log_collection = mongo_db[log_collection_name]
        self.test_configuration = test_configuration
        self.result_obj = {'testing_logs': [],
                           'training_logs': [],
                           'test_id': test_id,
                           'user_id': user_id,
                           'dataset_id': dataset_id,
                           'start_time_utc': str(datetime.datetime.utcnow()),
                           'test_configuration': self.test_configuration}

    def reset(self):
        """Reset start time and testing/training logs in result_obj
        """
        self.result_obj.update({'testing_logs': [],
                                'training_logs': [],
                                'start_time_utc': str(datetime.datetime.utcnow())})
        return

    def addLogRecord(self,
                     type: str,
                     record: dict):
        """Called during the testing loop to insert a pvt status
        record into MongoDB

        Args:
            type (str): Whether the record should be appended to the
                training or testing logs
            record (dict): the record to insert

        Raises:
            Exception: Thrown if the type provided is not supported
        """

        record_id = self.log_collection.insert_one(record).inserted_id
        if type == 'training':
            self.result_obj['training_logs'].append(record_id)
        elif type == 'testing':
            self.result_obj['testing_logs'].append(record_id)
        else:
            raise Exception(f'Unsupported record type {type} for \
                record {record}')

        return

    def saveResults(self,
                    final_results: dict):
        """Save a document in MongoDB, linking the result doc to
        the logs documents

        Args:
            final_results (dict): Information pertaining to the
                results of the test, to be stored in the results
                object for future use

        Returns:
            str: string of the ObjectId saved in MongoDB

        Example:
            .. code-block:: python

                uid = mongo_results.saveResults(final_state)
        """
        self.result_obj['end_time_utc'] = str(datetime.datetime.utcnow())
        self.result_obj['final_results'] = final_results
        # print(f'{self.result_obj = }')
        result_id = self.result_collection.insert_one(
            self.result_obj).inserted_id
        print(
            f'saved test results in {self.result_collection_name} as id: {result_id}')
        return str(result_id)

    def deleteResults(self):
        """Function used to remediate database in the event of a
        failed/aborted test

        Returns:
            dict: dict showing the deleted result record, if any
        """
        try:
            query_object = {'user_id': self.user_id,
                            'dataset_id': self.dataset_id,
                            'test_id': self.test_id
                            }
            delete_result = self.result_collection.find_one_and_delete(query_object,
                                                                       {'_id': False})

        except Exception as e:
            print(f'failed to delete results: {e}')
            return None

        if delete_result is not None:
            try:
                self.log_collection.delete_many(
                    {'_id': {'$in': delete_result['training_logs']}})
                self.log_collection.delete_many(
                    {'_id': {'$in': delete_result['testing_logs']}})

            except Exception as e:
                print(f'failed to delete log record: {e}')
                return None

        if self.result_obj['testing_logs']:
            self.log_collection.delete_many(
                {'_id': {'$in': self.result_obj['testing_logs']}})

        if self.result_obj['training_logs']:
            self.log_collection.delete_many(
                {'_id': {'$in': self.result_obj['training_logs']}})

        return {'status': 'deleted', 'record': delete_result}

    def retrieveResults(self):
        """Retreive test results from MongoDB based on user_id, dataset_id,
        and test_id

        Raises:
            Exception: If dataset master record is not found in database

        Returns:
            dict: Entire test result object
        """
        query_object = {'user_id': self.user_id,
                        'dataset_id': self.dataset_id,
                        'test_id': self.test_id
                        }
        self.result_obj = self.result_collection.find_one(query_object)

        if self.result_obj is None:
            raise Exception(f'Test not found based on {query_object=}')

        # retrieve successful
        training_ids = self.result_obj['training_logs']
        testing_ids = self.result_obj['testing_logs']
        training_logs = [item for item in self.log_collection.find(
            {'_id': {'$in': training_ids}})]
        testing_logs = [item for item in self.log_collection.find(
            {'_id': {'$in': testing_ids}})]

        self.result_obj['training_logs'] = training_logs
        self.result_obj['testing_logs'] = testing_logs

        return self.result_obj
